"""
Functionality of Amazon Web Services
"""

import boto3
from botocore.exceptions import ClientError

from .cfg import cfg
from .gen import generate


LINK = (
    f"https://{cfg('amazon.bucket')}.s3."
    f"{cfg('amazon.region')}.amazonaws.com/"
)


s3 = boto3.resource(
    's3',
    region_name=cfg('amazon.region'),
    aws_access_key_id=cfg('amazon.id'),
    aws_secret_access_key=cfg('amazon.secret'),
)
s3client = s3.meta.client


def upload_file(
    file,
    bucket=cfg('amazon.bucket'),
    directory=cfg('mode').lower(),
    file_type=None,
):
    """ Upload file """

    if not file_type and isinstance(file, str) and '.' in file:
        file_type = file.split('.')[-1]
    file_type = file_type.lower() if file_type else ''

    name = f"{directory}/{generate()}.{file_type}"

    if isinstance(file, str):
        handler = s3client.upload_file
    else:
        handler = s3client.upload_fileobj

    handler(
        file, bucket, name,
        ExtraArgs={'ACL': 'public-read'},
    )

    return LINK + name

def get_policy(bucket=cfg('amazon.bucket')):
    """ Get bucket policy """
    return s3client.get_bucket_policy(Bucket=bucket)


__all__ = (
    's3',
    's3client',
    'ClientError',
    'upload_file',
    'get_policy',
)



"""
Time functionality
"""

# TODO: Учитывать летнее / зимнее время в прошлых датах, которого теперь нет

import time
import datetime
# import pytz
import re

from .lang import get_form


MONTHS = {
    '01': ('январь', 'января', 'янв'),
    '02': ('февраль', 'февраля', 'февр', 'фев'),
    '03': ('март', 'марта', 'мар'),
    '04': ('апрель', 'апреля', 'апр'),
    '05': ('май', 'мая'),
    '06': ('июнь', 'июня', 'июн'),
    '07': ('июль', 'июля', 'июл'),
    '08': ('август', 'августа', 'авг'),
    '09': ('сентябрь', 'сентября', 'сент', 'сен'),
    '10': ('октябрь', 'октября', 'окт'),
    '11': ('ноябрь', 'ноября', 'нояб', 'ноя'),
    '12': ('декабрь', 'декабря', 'дек'),
}
DAYS_OF_WEEK = (
    'пн',
    'вт',
    'ср',
    'чт',
    'пт',
    'сб',
    'вс',
)


def get_time(data=time.time(), template='%d.%m.%Y %H:%M:%S', tz=0):
    """ Get time from timestamp """

    # TODO: smart TZ

    if isinstance(data, datetime.datetime):
        data = data.timestamp()

    return time.strftime(template, time.gmtime(data + tz * 3600))

# pylint: disable=too-many-branches,too-many-statements
def parse_time(data: str, tz=0):
    """ Parse time """

    # TODO: 16 year -> 2016 year

    data = data.lower()

    # Cut special characters
    data = re.sub(r'[^a-zа-я0-9:.]', '', data)

    # Cut the day of the week
    for day in DAYS_OF_WEEK:
        data = data.replace(day, '')

    data = data.strip()

    if len(data) < 4:
        return None

    for month_number, month_names in MONTHS.items():
        for month_name in month_names:
            if month_name in data:
                ind = data.index(month_name)
                data = data.replace(month_name, month_number)
                day = re.sub(r'[^0-9]', '', data[:ind])
                if day:
                    data = day + '.' + data[ind:]
                else:
                    data = '01.' + data[ind:]
                break
        else:
            continue
        break
    else:
        if len(data) != 8:
            proc = True
            if ':' in data:
                if len(re.sub(r'[^0-9]', '', data[:data.index(':')-2])) >= 6:
                    proc = False
            if proc:
                if '.' not in data:
                    data = '01.' + data
                if data.count('.') < 2:
                    data = '01.' + data

    if (
        ':' not in data
        and len(data) < 15
        and len(re.sub(r'[^0-9]', '', data)) <= 8
    ):
        data += '00:00:00'

    # Parse day
    if not data[1].isdigit():
        data = '0' + data
    if data[2] != '.':
        data = data[:2] + '.' + data[2:]

    # Parse month
    if data[5] != '.':
        data = data[:5] + '.' + data[5:]

    # Parse year
    data = data.replace('года', ' ')
    data = data.replace('год', ' ')
    data = data.replace('г.', ' ')
    if data[10] != ' ':
        data = data[:10] + ' ' + data[10:]

    # Timezone
    if 'msk' in data:
        data = data.replace('msk', '')
        tz_delta = 3
        # tz = pytz.timezone('Europe/Moscow')
    else:
        tz_delta = tz
        # tz = pytz.utc

    colon_count = data.count(':')
    if colon_count == 0 or colon_count > 2:
        return None
    if colon_count == 1:
        data += ':00'

    try:
        data = datetime.datetime.strptime(data, '%d.%m.%Y %H:%M:%S')
    except ValueError:
        return None

    data = data.replace(
        tzinfo=datetime.timezone(datetime.timedelta(hours=tz_delta))
    )

    return int(data.timestamp())

def format_delta(sec, short=False):
    """ Format time delta in words by seconds """

    if abs(sec) >= 259200: # 3 days
        time_def = round(sec / (24 * 60 * 60))
        if short:
            delta = f"{time_def}д"
        else:
            delta = f"{time_def} {get_form(time_def, ('день', 'дня', 'дней'))}"

    elif abs(sec) >= 10800: # 3 hours
        time_def = round(sec / (60 * 60))
        if short:
            delta = f"{time_def}ч"
        else:
            delta = f"{time_def} {get_form(time_def, ('час', 'часа', 'часов'))}"

    elif abs(sec) > 180: # 3 min
        time_def = round(sec / 60)
        if short:
            delta = f"{time_def}мин"
        else:
            delta = (
                f"{time_def}"
                f" {get_form(time_def, ('минута', 'минуты', 'минут'))}"
            )

    else:
        time_def = int(sec)
        if short:
            delta = f"{time_def}сек"
        else:
            delta = (
                f"{time_def}"
                f" {get_form(time_def, ('секунда', 'секунды', 'секунд'))}"
            )

    return delta



"""
Numbers functionality
"""

def is_float(value: str) -> bool:
    """ Check value for float """

    try:
        float(value)
    except (ValueError, TypeError):
        return False

    return True

def to_num(value) -> bool:
    """ Convert value to int or float """

    if value is None:
        return None

    if isinstance(value, str):
        value = float(value.strip())

    if not value % 1:
        value = int(value)

    return value



"""
Documents processing functionality
"""

import base64


def to_base64(image, mime='image/jpg'):
    """ Convert image to base64 """
    data = base64.b64encode(image.read())
    return f"data:{mime};base64,{data.decode('utf-8')}"



"""
Natural language processing functionality
"""

import re


TRANSLITERATION = {
    'а': 'a',
    'б': 'b',
    'в': 'v',
    'г': 'g',
    'д': 'd',
    'е': 'e',
    'ё': 'e',
    'ж': 'zh',
    'з': 'z',
    'и': 'i',
    'й': 'y',
    'к': 'k',
    'л': 'l',
    'м': 'm',
    'н': 'n',
    'о': 'o',
    'п': 'p',
    'р': 'r',
    'с': 's',
    'т': 't',
    'у': 'u',
    'ф': 'f',
    'х': 'kh',
    'ц': 'ts',
    'ч': 'ch',
    'ш': 'sh',
    'щ': 'shch',
    'ъ': '',
    'ы': 'y',
    'ь': '',
    'э': 'e',
    'ю': 'yu',
    'я': 'ya',
}


def get_form(count, variations):
    """ Get form of a noun with a number """

    count = abs(count)

    if count % 10 == 1 and count % 100 != 11:
        return variations[0]

    if count % 10 in (2, 3, 4) and count % 100 not in (12, 13, 14):
        return variations[1]

    return variations[2]

def transliterate(data, separator=' '):
    """ Transliterate RU → EN """

    if data is None:
        return ''

    data = ''.join(
        (
            i
            if 'a' <= i <= 'z' else
            TRANSLITERATION.get(i, separator)
        )
        for i in data.strip().lower()
    )
    data = re.sub(rf'{separator}{{2,}}', separator, data)

    return data

def to_letters(data, separator=''):
    """ To letters & numbers """

    if data is None:
        return ''

    data = re.sub('[^a-zа-я0-9]+', ' ' if separator else '' , data.lower())
    if separator:
        data = data.strip().replace(' ', separator)

    return data



"""
Codes generation functionality
"""

import string
import random


SYMBOLS = string.digits + string.ascii_letters
SPEC_SYMBOLS = '!@#$%&*-+=,./?|]~'


def generate(length: int = 32) -> str:
    """ Token / code generation """
    return ''.join(random.choice(SYMBOLS) for _ in range(length))

def generate_id(length: int = 8) -> int:
    """ ID generation """
    number = ''.join(random.choice(string.digits) for _ in range(length-1))
    number = random.choice(string.digits[1:]) + number
    return int(number)

def generate_password(length: int = 8) -> str:
    """ Password generation """

    spec_symbols = ''.join(
        random.choice(SPEC_SYMBOLS) for _ in range(length//3)
    )
    digits = ''.join(
        random.choice(string.digits) for _ in range(length//3)
    )
    letters = ''.join(
        random.choice(string.ascii_letters) for _ in range(
            length - 2 * (length//3)
        )
    )

    data = list(spec_symbols + digits + letters)
    random.shuffle(data)
    data = ''.join(data)

    return data



"""
Database ciphers
"""

from .cfg import cfg


# NOTE: ISO 639-1
LOCALES = (
    'en',
    'ru',
    'es',
)
FLAGS = (
    '🇬🇧',
    '🇷🇺',
    '🇪🇸',
)

CURRENCIES = {
    'USD': '$',
    'RUB': '₽',
    'EUR': '€',
    'TRY': '₺',
    'UAH': '₴',
    'THB': '฿',
}

NETWORKS = (
    '', # Console
    'web', # Web-interface
    'tg', # Telegram
    'vk', # VKontakte
    'g', # Google
    'fb', # Facebook
    'a', # Apple
    'in', # LinkedIn
    'ig', # Instagram
)

STATUSES = (
    'removed',
    'disabled',
    'active',
)
USER_STATUSES = (
    'removed', # deleted # not specified # Does not exist
    'blocked', # archive # Does not have access to resources
    'guest', # normal
    'authorized', # registered # confirmed # Save personal data & progress
    'editor', # curator # View reviews
    'verified', # Delete reviews
    'moderator', # Block users
    'admin', # Delete posts
    'owner', # Can't be blocked
)


default_locale = cfg('locale', 0)


def get_network(code):
    """ Get network code by cipher """

    if code is None:
        return 0

    if code in NETWORKS:
        return NETWORKS.index(code)

    if code in range(len(LOCALES)):
        return code

    return 0

def get_locale(code):
    """ Get language code by cipher """

    if code is None:
        return default_locale

    if code in LOCALES:
        return LOCALES.index(code)

    if code in range(len(LOCALES)):
        return code

    return default_locale

def get_flag(code):
    """ Get flag by language """
    return FLAGS[get_locale(code)]



"""
Checking functionality
"""

def fake_phone(value: str) -> bool:
    """ Check a phone for a test format """

    if value is None:
        return False

    value = str(value)

    return any(
        fake in value
        for fake in (
            '0000', '1111', '2222', '3333', '4444', '5555', '6666', '7777',
            '8888', '9999', '1234', '2345', '3456', '4567', '5678', '6789',
            '9876', '8765', '7654', '6543', '5432', '4321',
        )
    )

def fake_login(value: str) -> bool:
    """ Check a login / name / mail for a test format """

    if value is None:
        return False

    value = value.lower()

    return any(
        fake in value
        for fake in (
            'test', 'тест', 'check',
            'asd', 'qwe', 'rty', 'sdf', 'sfg', 'sfd', 'hgf', 'gfd',
            'qaz', 'wsx', 'edc', 'rfv',
            '111', '123',
            'ыва', 'фыв', 'йцу', 'орп',
        )
    )



"""
Functionality of getting configuration
"""

import os
import json

from dotenv import load_dotenv


if os.path.isfile('sets.json'):
    with open('sets.json', 'r', encoding='utf-8') as file:
        sets = json.loads(file.read())
else:
    sets = {}

if os.path.isfile('.env'):
    load_dotenv()


def cfg(name, default=None):
    """ Get config value by key """

    keys = name.split('.')
    data = sets

    for key in keys:
        if key not in data:
            break
        data = data[key]
    else:
        return data

    name = name.replace('.', '_').upper()
    value = os.getenv(name, default)

    if value:
        try:
            value = json.loads(value)
        except (json.decoder.JSONDecodeError, TypeError):
            pass

    return value


def set_cfg(name, value):
    """ Set config value """

    array_name = name.split('.')
    dictionary = {}
    tmp_dict = {}

    if len(array_name) == 1:
        sets[name] = value
        return

    index = len(array_name) - 1
    dictionary[array_name[index]] = value
    index -= 1
    while index > 0:
        if array_name[index] in sets:
            tmp_dict[array_name[index]].append(dictionary)
        else:
            tmp_dict[array_name[index]] = dictionary
        dictionary = tmp_dict
        tmp_dict = {}
        index -= 1

    if array_name[0] not in sets:
        sets[array_name[0]] = dictionary
    else:
        sets[array_name[0]].update(dictionary)
